#ifndef OPENBRAIN_H
#define OPENBRAIN_H

#include <QtGui/QAction>
#include <QtGui/QGraphicsLinearLayout>
#include <QtGui/QGraphicsWidget>

#include <Plasma/PopupApplet>
#include <Plasma/IconWidget>
#include <Plasma/Label>
#include <Plasma/LineEdit>

#include <iostream>
#include <string>

#include "openbrainsettings.h"
#include "aimlparser.h"

using namespace std;

class OpenBrain : public Plasma::PopupApplet
{
  Q_OBJECT

  public:

    OpenBrain(QObject *parent, const QVariantList &args);
    ~OpenBrain();

    void init();
    void createConfigurationInterface(KConfigDialog *parent);

    QGraphicsWidget *graphicsWidget();
    
  public slots:
    void configAccepted();
    void readConfig();
    void writeConfig();
    void displayHome();
    void loadBrain();
    void returnPressed();
    void updateLoadingPage(const QString&);
    void doneLoading();

  private:
    void setupInitialProperties();
    void setupConnections();
    void setupLayout();

    QAction *m_goAction;
    QAction *m_homeAction;

    QGraphicsLinearLayout *m_layout;
    QGraphicsLinearLayout *m_inputLayout;
    QGraphicsWidget *main_widget;

    QString executeCommand(const QString&);
    QString thinkOf(QString s);

    bool cleartext;

    int m_horizontalScrollValue;
    int m_verticalScrollValue;

    Plasma::IconWidget *m_go;
    Plasma::IconWidget *m_home;
    Plasma::LineEdit *m_inputLineEdit;
    Plasma::Label *m_response;

    QString m_aimlset;
    QFont m_font;
    QColor m_color;
    bool m_spellcheck;
    
    OpenBrainSettings *m_settings;
    
    AIMLParser *parser;
};

K_EXPORT_PLASMA_APPLET(openbrain, OpenBrain)

#endif
